from __future__ import annotations

import logging
import numpy as np
import time
import wgpu
import wgpu.backends.rs # noqa: F401, Select Rust backend

from . import config
from . import canvas
from . import colormap
from . import multiresolution_sph, sph
from . import colorbar
from . import text
from . import scalebar
from . import loader
from . import util

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Visualizer:
    colormap_name = config.DEFAULT_COLORMAP
    colorbar_aspect_ratio = config.COLORBAR_ASPECT_RATIO

    show_status = True
    def __init__(self, data_loader_class = loader.TestDataLoader, data_loader_args = ()):


        self.canvas = canvas.VisualizerCanvas(visualizer=self, title="topsy")
        self.adapter: wgpu.GPUAdapter = wgpu.request_adapter(canvas=self.canvas, power_preference="high-performance")
        self.device: wgpu.GPUDevice = self.adapter.request_device()
        self.context: wgpu.GPUCanvasContext = self.canvas.get_context()
        self.canvas_format = self.context.get_preferred_format(self.adapter)

        self._n_sph_channels = 1
        self._recent_frame_times = []
        if self.canvas_format.endswith("-srgb"):
            # matplotlib colours aren't srgb. It might be better to convert
            # but for now, just stop the canvas being srgb
            self.canvas_format = self.canvas_format[:-5]

        self.context.configure(device=self.device, format=self.canvas_format)

        self._render_resolution = 1024

        self.render_texture: wgpu.GPUTexture = self.device.create_texture(
            size=(self._render_resolution, self._render_resolution, 1),
            usage=wgpu.TextureUsage.RENDER_ATTACHMENT |
                  wgpu.TextureUsage.TEXTURE_BINDING |
                  wgpu.TextureUsage.COPY_SRC,
            format=wgpu.TextureFormat.rg32float,
            label="sph_render_texture",
        )

        self.data_loader = data_loader_class(self.device, *data_loader_args)

        self._colormap = colormap.Colormap(self, weighted_average = False)
        self._sph = sph.SPH(self, self.render_texture)
        #self._sph = multiresolution_sph.MultiresolutionSPH(self, self.render_texture)

        self._last_status_update = 0.0

        self._colorbar = colorbar.ColorbarOverlay(self, 0.0, 1.0, self.colormap_name, "TODO")
        self._scalebar = scalebar.ScalebarOverlay(self)
        self._status = text.TextOverlay(self, "topsy", (-0.9, 0.9), 80, color=(1, 1, 1, 1))
        self._display_fullres_render_status = False # when True, customises info text to refer to full-res render

        self._render_timer = util.TimeGpuOperation(self.device)

        self.vmin_vmax_is_set = False

        self.invalidate()



    def invalidate(self):
        self.canvas.request_draw(self.draw)

    def rotate(self, dx, dy):
        dx_rotation_matrix = self._x_rotation_matrix(dx*0.01)
        dy_rotation_matrix = self._y_rotation_matrix(dy*0.01)
        self._sph.rotation_matrix = dx_rotation_matrix @ dy_rotation_matrix @ self._sph.rotation_matrix
        self.invalidate()

    def reset_view(self):
        self._sph.rotation_matrix = np.eye(3)
        self.scale = config.DEFAULT_SCALE

    @property
    def scale(self):
        """Return the scalefactor from kpc to viewport coordinates. Viewport will therefore be 2*scale wide."""
        return self._sph.scale
    @scale.setter
    def scale(self, value):
        self._sph.scale = value
        self.invalidate()

    @property
    def quantity_name(self):
        """The name of the quantity being visualised, or None if density projection."""
        return self.data_loader.quantity_name

    @property
    def averaging(self):
        """True if the quantity being visualised is a weighted average, False if it is a mass projection."""
        return self.data_loader.quantity_name is not None

    @quantity_name.setter
    def quantity_name(self, value):
        self.data_loader.quantity_name = value
        self.vmin_vmax_is_set = False
        self._colormap = colormap.Colormap(self, weighted_average = value is not None)
        self._colorbar = colorbar.ColorbarOverlay(self, 0.0, 1.0, self.colormap_name,
                                                  r"$\log_{10}$ "+self.data_loader.get_quantity_label())
        self.invalidate()

    @staticmethod
    def _y_rotation_matrix(angle):
        return np.array([[1, 0, 0],
                         [0, np.cos(angle), -np.sin(angle)],
                         [0, np.sin(angle), np.cos(angle)]])

    @staticmethod
    def _x_rotation_matrix(angle):
        return np.array([[np.cos(angle), 0, np.sin(angle)],
                         [0, 1, 0],
                         [-np.sin(angle), 0, np.cos(angle)]])

    def _check_whether_inactive(self):
        if time.time()-self._last_lores_draw_time>config.FULL_RESOLUTION_RENDER_AFTER*0.999:
            self._sph.downsample_factor = 1
            self._last_lores_draw_time = np.inf
            self._display_fullres_render_status = True
            self.invalidate()



    def draw(self):

        ce_label = "sph_render"
        # labelling this is useful for understanding performance in macos instruments
        if self._sph.downsample_factor>1:
            ce_label += f"_ds{self._sph.downsample_factor:d}"
        else:
            ce_label += "_fullres"

        command_encoder : wgpu.GPUCommandEncoder = self.device.create_command_encoder(label=ce_label)
        self._sph.encode_render_pass(command_encoder)

        with self._render_timer:
            self.device.queue.submit([command_encoder.finish()])

        # in principle, we can often use the same command encoder for the drawing into the final
        # buffer, but when vmin/vmax needs to be set, we need to render the image first.
        # So we just always use a new command encoder for the final render. It is unclear whether
        # this has any performance impact.

        if not self.vmin_vmax_is_set:
            logger.info("Setting vmin/vmax")
            self._colormap.set_vmin_vmax()
            self.vmin_vmax_is_set = True
            self._colorbar.vmin = self._colormap.vmin
            self._colorbar.vmax = self._colormap.vmax
            self._colorbar.update()

        command_encoder = self.device.create_command_encoder(label="render_to_screen")
        self._colormap.encode_render_pass(command_encoder)
        self._colorbar.encode_render_pass(command_encoder)
        self._scalebar.encode_render_pass(command_encoder)

        if self.show_status:
            self._update_and_display_status(command_encoder)

        self.device.queue.submit([command_encoder.finish()])

        if self._sph.downsample_factor>1:
            self._last_lores_draw_time = time.time()
            wgpu.gui.auto.call_later(config.FULL_RESOLUTION_RENDER_AFTER, self._check_whether_inactive)
        elif self._render_timer.last_duration>1/config.TARGET_FPS and self._sph.downsample_factor==1:
            # this will affect the NEXT frame, not this one!
            self._sph.downsample_factor = int(np.floor(float(config.TARGET_FPS)*self._render_timer.last_duration))


    def _update_and_display_status(self, command_encoder):
        now = time.time()
        if now - self._last_status_update > config.STATUS_LINE_UPDATE_INTERVAL:
            self._last_status_update = now
            self._status.text = f"${1.0 / self._render_timer.running_mean_duration:.0f}$ fps"
            if self._sph.downsample_factor > 1:
                self._status.text += f", downsample={self._sph.downsample_factor:d}"
            if self._display_fullres_render_status:
                self._status.text = f"Full-res render took {self._render_timer.last_duration:.2f} s"
                self._display_fullres_render_status = False
            self._status.update()

        self._status.encode_render_pass(command_encoder)

    def get_rendered_image(self) -> np.ndarray:
        im = self.device.queue.read_texture({'texture':self.render_texture, 'origin':(0, 0, 0)},
                                            {'bytes_per_row':8*self._render_resolution},
                                            (self._render_resolution, self._render_resolution, 1))
        np_im = np.frombuffer(im, dtype=np.float32).reshape((self._render_resolution, self._render_resolution, 2))
        if self.averaging:
            im = np_im[:,:,1]/np_im[:,:,0]
        else:
            im = np_im[:,:,0]
        return im

    def save(self, filename='output.pdf'):
        image = self.get_rendered_image()
        import pylab as p
        fig = p.figure()
        p.clf()
        p.set_cmap(self.colormap_name)
        extent = np.array([-1., 1., -1., 1.])*self.scale
        if self._colormap.log_scale:
            image = np.log10(image)

        p.imshow(image,
                 vmin=self._colormap.vmin,
                 vmax=self._colormap.vmax,
                 extent=extent)
        p.xlabel("$x$/kpc")
        p.colorbar().set_label(self.colorbar_label)
        p.savefig(filename)
        p.close(fig)

    def run(self):
        wgpu.gui.auto.run()



"""topsy - An astrophysics simulation visualization package based on OpenGL, using pynbody for reading data"""

__version__ = "0.2.0"

import logging
import sys
import argparse
import matplotlib
import pynbody
import wgpu.gui.jupyter

from . import config, visualizer, loader

matplotlib.use("Agg")


def parse_args():
    """Create arguments and kwargs to pass to the visualizer, from sys.argv"""

    # create the argument parser, add arguments for filename, resolution, and colormap, and parse the arguments
    argparser = argparse.ArgumentParser(description="Visualize an astrophysics simulation")

    argparser.add_argument("filename", help="Specify path to a simulation file to be visualized")
    argparser.add_argument("--resolution", "-r", help="Specify the resolution of the visualization",
                           default=config.DEFAULT_RESOLUTION, type=int)
    argparser.add_argument("--colormap", "-m", help="Specify the matplotlib colormap to be used",
                           default=config.DEFAULT_COLORMAP, type=str)
    argparser.add_argument("--particle", "-p", help="Specify the particle type to visualise",
                            default="dm", type=str)
    argparser.add_argument("--center", "-c", help="Specify the centering method: 'halo-<N>', 'all', 'zoom' or 'none'",
                           default="halo-1", type=str)
    argparser.add_argument("--quantity", "-q", help="Specify a quantity to render instead of density",
                           default=None, type=str)

    return argparser.parse_args()

def setup_logging():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(ch)

def main():
    setup_logging()
    args = parse_args()

    vis = visualizer.Visualizer(data_loader_class=loader.PynbodyDataLoader,
                                data_loader_args=(args.filename, args.center,
                                                  args.particle))
    vis.quantity_name = args.quantity
    vis.run()

def topsy(snapshot: pynbody.snapshot.SimSnap, quantity: str | None = None):
    vis = visualizer.Visualizer(data_loader_class=loader.PynbodyDataInMemory,
                                data_loader_args=(snapshot,))
    vis.quantity_name = quantity
    if isinstance(vis.canvas, wgpu.gui.jupyter.JupyterWgpuCanvas):
        return vis
    else:
        vis.run()


from __future__ import annotations

import numpy as np
import wgpu.gui.jupyter

from wgpu.gui.auto import WgpuCanvas

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

class VisualizerCanvas(WgpuCanvas):
    def __init__(self, *args, **kwargs):
        self._visualizer : Visualizer = kwargs.pop("visualizer")

        self._last_x = 0
        self._last_y = 0
        # The below are dummy values that will be updated by the initial resize event
        self.width_physical, self.height_physical = 640,480
        self.pixel_ratio = 1

        super().__init__(*args, **kwargs)

    def handle_event(self, event):
        if event['event_type']=='pointer_move':
            if len(event['buttons'])>0:
                self.drag(event['x']-self._last_x, event['y']-self._last_y)
            self._last_x = event['x']
            self._last_y = event['y']
        elif event['event_type']=='wheel':
            self.mouse_wheel(event['dx'], event['dy'])
        elif event['event_type']=='key_up':
            self.key_up(event['key'])
        elif event['event_type']=='resize':
            self.resize(event['width'], event['height'], event['pixel_ratio'])
        else:
            pass
        super().handle_event(event)

    def drag(self, dx, dy):
        self._visualizer.rotate(dx, dy)

    def key_up(self, key):
        if key=='s':
            self._visualizer.save()
        elif key=='r':
            self._visualizer.vmin_vmax_is_set = False
            self._visualizer.invalidate()
        elif key=='h':
            self._visualizer.reset_view()

    def mouse_wheel(self, delta_x, delta_y):
        if isinstance(self, wgpu.gui.jupyter.JupyterWgpuCanvas):
            # scroll events are much smaller from the web browser, for
            # some reason, compared with native windowing
            delta_y *= 10
            delta_x *= 10

        self._visualizer.scale*=np.exp(delta_y/1000)

    def resize(self, width, height, pixel_ratio=1):
        self.width_physical = width*pixel_ratio
        self.height_physical = height*pixel_ratio
        self.pixel_ratio = pixel_ratio


import numpy as np
import wgpu
import logging
import pynbody
import pickle 

# ABC support:
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)
class AbstractDataLoader(ABC):
    def __init__(self, device: wgpu.GPUDevice):
        self._device = device
        self.quantity_name = None
        self._quantity_buffer_is_for_name = None
    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def get_positions(self):
        pass

    @abstractmethod
    def get_smooth(self):
        pass

    @abstractmethod
    def get_mass(self):
        pass

    @abstractmethod
    def get_named_quantity(self, name):
        pass

    def get_pos_smooth(self):
        pos_smooth = np.empty((len(self), 4), dtype=np.float32)
        pos_smooth[:, :3] = self.get_positions()
        pos_smooth[:, 3] = self.get_smooth()
        return pos_smooth

    def get_pos_smooth_buffer(self):
        if not hasattr(self, "_pos_smooth_buffer"):
            logger.info("Creating position+smoothing buffer")
            data = self.get_pos_smooth()
            self._pos_smooth_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
        return self._pos_smooth_buffer

    def get_mass_buffer(self):
        if not hasattr(self, "_mass_buffer"):
            logger.info("Creating mass buffer")
            data = self.get_mass()
            self._mass_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
        return self._mass_buffer

    def get_quantity_buffer(self):
        if self.quantity_name is None:
            return self.get_mass_buffer()
        elif self._quantity_buffer_is_for_name != self.quantity_name:
            logger.info(f"Creating {self.quantity_name} buffer")
            data = self.get_named_quantity(self.quantity_name)
            self._named_quantity_buffer = self._device.create_buffer_with_data(
                data=data,
                usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.UNIFORM)
            self._quantity_buffer_is_for_name = self.quantity_name
        return self._named_quantity_buffer


class PynbodyDataInMemory(AbstractDataLoader):
    """Base class for data loaders that use pynbody."""

    def __init__(self, device: wgpu.GPUDevice, snapshot: pynbody.snapshot.SimSnap):
        super().__init__(device)

        self.snapshot = snapshot

        # randomize order to avoid artifacts when downsampling number of particles on display
        self._random_order = np.random.permutation(len(self.snapshot))

    def get_positions(self):
        return self.snapshot['pos'].astype(np.float32)[self._random_order]

    def get_smooth(self):
        return self.snapshot['smooth'].astype(np.float32)[self._random_order]

    def get_mass(self):
        return self.snapshot['mass'].astype(np.float32)[self._random_order]

    def get_named_quantity(self, name):
        return self.snapshot[name].astype(np.float32)[self._random_order]

    def get_quantity_label(self):
        if self.quantity_name is None:
            return r"density / $M_{\odot} / \mathrm{kpc}^2$"
        else:
            lunit = self.snapshot[self.quantity_name].units.latex()
            if lunit != "":
                lunit = "$/" + lunit + "$"
            return self.quantity_name + lunit

    def __len__(self):
        return len(self.snapshot)

class PynbodyDataLoader(PynbodyDataInMemory):
    """Literal data loader for pynbody (starts from just a filename)"""
    def __init__(self, device: wgpu.GPUDevice, filename: str, center: str, particle: str):

        logger.info(f"Data filename = {filename}, center = {center}, particle = {particle}")
        snapshot = pynbody.load(filename)
        snapshot.physical_units()
        self.filename = filename

        snapshot = snapshot[pynbody.family.get_family(particle)]

        super().__init__(device, snapshot)

        self._perform_centering(center)
        self._perform_smoothing()

    def _perform_centering(self, center):
        logger.info("Performing centering...")
        if center.startswith("halo-"):
            halo_number = int(center[5:])
            h = self.snapshot.ancestor.halos()
            pynbody.analysis.halo.center(h[halo_number])

        elif center == 'zoom':
            f_dm = self.snapshot.ancestor.dm
            pynbody.analysis.halo.center(f_dm[f_dm['mass'] < 1.01 * f_dm['mass'].min()])
        elif center == 'all':
            pynbody.analysis.halo.center(self.snapshot)
        elif center == 'none':
            pass
        else:
            raise ValueError("Unknown centering type")

    def _perform_smoothing(self):
        try:
            logger.info("Looking for cached smoothing/density data...")
            smooth = pickle.load(open(self.filename+'-topsy-smooth.pkl', 'rb'))
            if len(smooth) == len(self.snapshot):
                self.snapshot['smooth'] = smooth
            else:
                raise ValueError("Incorrect number of particles in cached smoothing data")
            logger.info("...success!")

            rho = pickle.load(open(self.filename+'-topsy-rho.pkl', 'rb'))
            if len(rho) == len(self.snapshot):
                self.snapshot['rho'] = rho
            else:
                raise ValueError("Incorrect number of particles in cached density data")
        except:
            logger.info("Generating smoothing/density data - this can take a while but will be cached for future runs")
            pickle.dump(self.snapshot['smooth'], open(self.filename+'-topsy-smooth.pkl', 'wb'))
            pickle.dump(self.snapshot['rho'], open(self.filename+'-topsy-rho.pkl', 'wb'))



class TestDataLoader(AbstractDataLoader):
    def __init__(self, device: wgpu.GPUDevice, n_particles: int = int(5e6)):
        self._n_particles = n_particles
        self._gmm_weights = [0.5, 0.25, 0.25] # should sum to 1
        self._gmm_means = np.array([[0.0, 0.0, 0.0],[0.0, 0.0, 0.0],[0.6,0.0,0.0]])
        self._gmm_std = np.array([[0.2, 0.2, 0.2], [0.4, 0.02, 0.4], [0.1, 0.1, 0.1]])

        self._gmm_weights = [1.0]
        self._gmm_means = np.array([[0.0, 0.0, 0.0]])
        self._gmm_std = np.array([[0.2, 0.2, 0.2]])

        self._gmm_pos = self._generate_samples()
        self._gmm_den = self._evaluate_density(self._gmm_pos)
        super().__init__(device)
    def __len__(self):
        return self._n_particles

    def _evaluate_density(self, pos):
        # simple gaussian mixture model, density of particles per unit volume
        den = np.zeros(len(pos))
        for i in range(len(self._gmm_weights)):
            den += self._gmm_weights[i] * np.exp(-np.sum((pos - self._gmm_means[i])**2, axis=1)) \
                                                 / ((2 * np.pi)**1.5 * np.prod(self._gmm_std[i]))
        return den*self._n_particles

    def _generate_samples(self):
        # simple gaussian mixture model
        pos = np.empty((self._n_particles, 3))
        offset = 0
        for i in range(len(self._gmm_weights)):
            cpt_len = int(self._n_particles*self._gmm_weights[i])
            pos[offset:offset+cpt_len] = \
                np.random.normal(size=(cpt_len, 3), scale=1.0).astype(np.float32) * self._gmm_std[np.newaxis,i,:] + self._gmm_means[i]
            offset += cpt_len
        assert offset == self._n_particles
        return pos

    def get_positions(self):

        return self._gmm_pos

    def get_smooth(self):
        return 10./self._gmm_den**0.333333
        #return np.random.uniform(0.01, 0.05, size=(self._n_particles)).astype(np.float32)

    def get_mass(self):
        return np.random.uniform(0.01, 1.0, size=(self._n_particles)).astype(np.float32)*1e-8